/*
 * VisInfo.h
 *
 * Definition of the VisualInfo structure and (pseudo) prototype of the
 * XfCreateVisual() function.
 *
 *
 */

#ifndef _XF_VISINFO_H
#define _XF_VISINFO_H

/* Values for the type selector in the Visual struct */
#define XfPixmapVisual 0
#define XfTextVisual 1
#define XfXImageVisual 2
#define XfTiledVisual 3
#define XfStippledVisual 4
#define XfOpaqueStippledVisual 5
#define XfSolidVisual 6
#define XfOutlineVisual 7

/* A visual element comprised of normal text */
struct TextVisual
{
  char text[256];               /* The actual text displayed */
  Font font;                    /* The font to use in displaying */
  int align;                    /* 0 is center, 1 is left, 2 is right */
};

/* A visual element comprised of a bitmap or pixmap */
struct PixmapVisual
{
  int depth;                    /* Bit depth of image */
  Pixmap map;                   /* Pixmap data */
};

/* The structure combining image and text visuals into one treatment */
struct VisualInfo
{
  int x_pos, y_pos;             /* position relative to button's X & Y */
  int height, width;            /* Height and width of this little piece */
  unsigned long foreground;     /* Foreground color for this element */
  unsigned long background;     /* Background for element */
  int vtype;                    /* Type of element-- text or image */
  union Visual                  /* The needed data for displaying */
    {
      struct TextVisual t_vis;
      struct PixmapVisual p_vis;
      XImage* i_vis;
    } visual;
  struct VisualInfo* next;      /* Handle to the next bit of Visual info */
};

/* The routine to create the above mess */
struct VisualInfo* XfCreateVisual();

#endif /* _XF_VISINFO_H */

